import pandas as pd
import numpy as np
import sciris as sc
import numba as nb
from starsim.utils import INT_NAN
from starsim.distributions import ScipyDistribution
from starsim.utils import warn
from numpy.lib.mixins import NDArrayOperatorsMixin  # Inherit from this to automatically gain operators like +, -, ==, <, etc.
from scipy.stats._distn_infrastructure import rv_frozen

__all__ = ['State', 'DynamicView']



class FusedArray(NDArrayOperatorsMixin):
    # This is a class that allows indexing by UID but does not support dynamic growth
    # It's kind of like a Pandas series but one that only supports a monotonically increasing
    # unique integer index, and that we can customize and optimize indexing for.
    #
    # We explictly do NOT support slicing (except for `[:]`), as these arrays are indexed by UID and slicing
    # by UID can be confusing/ambiguous when there are missing values. Indexing with a list/array returns
    # another FusedArray instance which enables chained filtering

    __slots__ = ('values','_uid_map', 'uid')

    def __init__(self, values, uid, uid_map=None):

        self.values = values
        self.uid = uid

        if uid_map is None and uid is not None:
            # Construct a local UID map as opposed to using a shared one (i.e., the one for all agents contained in the People instance)
            self.uid_map = np.full(np.max(uid) + 1, fill_value=INT_NAN, dtype=int)
            self.uid_map[uid] = np.arange(len(uid))
        else:
            self._uid_map = uid_map

    def __repr__(self):
        # TODO - optimize? Don't really need to create a dataframe just to print it, but on the other hand, it's fast enough and very easy
        df = pd.DataFrame(self.values.T, index=self.uid, columns=['Quantity'])
        df.index.name = 'UID'
        return df.__repr__()

    @property
    def dtype(self):
        return self.values.dtype

    @staticmethod
    @nb.njit
    def _get_vals_uids(vals, key, uid_map):
        """
        Extract values from a collection of UIDs

        This function is used to retrieve values based on UID. As indexing a FusedArray returns a new FusedArray,
        this method also populates the new UID map for use in the subsequently created FusedArray, avoiding the
        need to re-compute it separately.

        :param vals: A 1D np.ndarray containing the values
        :param key: A 1D np.ndnarray of integers containing the UIDs to query
        :param uid_map: A 1D np.ndarray of integers mapping UID to array position in ``vals``
        :return: A tuple of (values, uids, new_uid_map) suitable for passing into the FusedArray constructor
        """
        out = np.empty(len(key), dtype=vals.dtype)
        new_uid_map = np.full(uid_map.shape[0], fill_value=INT_NAN, dtype=np.int64)

        for i in range(len(key)):
            idx = uid_map[key[i]]
            if idx == INT_NAN:
                raise IndexError('UID not present in array')
            out[i] = vals[idx]
            new_uid_map[key[i]] = i
        return out, key, new_uid_map

    @staticmethod
    @nb.njit
    def _set_vals_uids_multiple(vals, key, uid_map, value):
        """
        Insert an array of values based on UID

        :param vals: A reference to a 1D np.ndarray in which to insert the values
        :param key: A 1D np.ndnarray of integers containing the UIDs to add values for
        :param uid_map:  A 1D np.ndarray of integers mapping UID to array position in ``vals``
        :param value: A 1D np.ndarray the same length as ``key`` containing values to insert
        :return:
        """

        for i in range(len(key)):
            if key[i] >= len(uid_map):
                raise IndexError('UID not present in array (requested UID is larger than the maximum UID in use)')
            idx = uid_map[key[i]]
            if idx == INT_NAN:
                raise IndexError('UID not present in array')
            elif idx >= len(vals):
                raise Exception(f'Attempted to write to a non-existant index - this can happen if attempting to write to new entries that have not yet been allocated via grow()')
            vals[idx] = value[i]

    @staticmethod
    @nb.njit
    def _set_vals_uids_single(vals, key, uid_map, value):
        """
        Insert a single value into multiple UIDs

        :param vals: A reference to a 1D np.ndarray in which to insert the values
        :param key: A 1D np.ndnarray of integers containing the UIDs to add values for
        :param uid_map:  A 1D np.ndarray of integers mapping UID to array position in ``vals``
        :param value: A scalar value to insert at every position specified by ``key``
        :return:
        """
        for i in range(len(key)):
            if key[i] >= len(uid_map):
                raise IndexError('UID not present in array (requested UID is larger than the maximum UID in use)')
            idx = uid_map[key[i]]
            if idx == INT_NAN:
                raise IndexError('UID not present in array')
            elif idx >= len(vals):
                raise Exception('Attempted to write to a non-existant index - this can happen if attempting to write to new entries that have not yet been allocated via grow()')
            vals[idx] = value

    def __getitem__(self, key):
        try:
            if isinstance(key, (int, np.integer)):
                # Handle getting a single item by UID
                return self.values[self._uid_map[key]]
            elif isinstance(key, (np.ndarray, FusedArray, DynamicView)):
                if key.dtype.kind == 'b':
                    # Handle accessing items with a logical array. Surprisingly, it seems faster to use nonzero() to convert
                    # it to indices first. Also, the pure Python implementation is difficult to improve upon using numba
                    mapped_key = key.__array__().nonzero()[0]
                    uids = self.uid.__array__()[mapped_key]
                    new_uid_map = np.full(len(self._uid_map), fill_value=INT_NAN, dtype=int)
                    new_uid_map[uids] = np.arange(len(uids))
                    values = self.values[mapped_key]
                else:
                    # Access items by an array of integers. We do get a decent performance boost from using numba here
                    values, uids, new_uid_map = self._get_vals_uids(self.values, key.__array__(), self._uid_map.__array__())
            elif isinstance(key, slice):
                if key.start is None and key.stop is None and key.step is None:
                    return sc.dcp(self)
                else:
                    raise Exception('Slicing not supported - slice the .values attribute by index instead e.g., x.values[0:5], not x[0:5]')
            else:
                # This branch is specifically handling the user passing in a list of integers instead of an array, therefore
                # we need an additional conversion to an array first using np.fromiter to improve numba performance
                values, uids, new_uid_map = self._get_vals_uids(self.values, np.fromiter(key, dtype=int), self._uid_map.__array__())
            return FusedArray(values=values, uid=uids, uid_map=new_uid_map)
        except IndexError as e:
            if str(INT_NAN) in str(e):
                raise IndexError(f'UID not present in array')
            else:
                raise e

    def __setitem__(self, key, value):
        # nb. the use of .__array__() calls is to access the array interface and thereby treat both np.ndarray and DynamicView instances
        # in the same way without needing an additional type check. This is also why the FusedArray.dtype property is defined. Noting
        # that for a State, the uid_map is a dynamic view attached to the People, but after an indexing operation, it will be a bare
        # FusedArray that has an ordinary numpy array as the uid_map
        try:
            if isinstance(key, (int, np.integer)):
                return self.values.__setitem__(self._uid_map[key], value)
            elif isinstance(key, (np.ndarray, FusedArray)):
                if key.dtype.kind == 'b':
                    self.values.__setitem__(key.__array__().nonzero()[0], value)
                else:
                    if isinstance(value, (np.ndarray, FusedArray)):
                        return self._set_vals_uids_multiple(self.values, key, self._uid_map.__array__(), value.__array__())
                    else:
                        return self._set_vals_uids_single(self.values, key, self._uid_map.__array__(), value)
            elif isinstance(key, slice):
                if key.start is None and key.stop is None and key.step is None:
                    return self.values.__setitem__(key, value)
                else:
                    raise Exception('Slicing not supported - slice the .values attribute by index instead e.g., x.values[0:5], not x[0:5]')
            else:
                if isinstance(value, (np.ndarray, FusedArray)):
                    return self._set_vals_uids_multiple(self.values, np.fromiter(key, dtype=int), self._uid_map.__array__(), value.__array__())
                else:
                    return self._set_vals_uids_single(self.values, np.fromiter(key, dtype=int), self._uid_map.__array__(), value)
        except IndexError as e:
            if str(INT_NAN) in str(e):
                raise IndexError(f'UID not present in array')
            else:
                raise e

    # Make it behave like a regular array mostly
    def __len__(self):
        return len(self.values)

    def __contains__(self, *args, **kwargs):
        return self.values.__contains__(*args, **kwargs)

    def astype(self, *args, **kwargs):
        return FusedArray(values=self.values.astype(*args, **kwargs), uid=self.uid, uid_map=self._uid_map)

    def sum(self, *args, **kwargs):
        return self.values.sum(*args, **kwargs)

    def mean(self, *args, **kwargs):
        return self.values.mean(*args, **kwargs)

    def any(self, *args, **kwargs):
        return self.values.any(*args, **kwargs)

    def all(self, *args, **kwargs):
        return self.values.all(*args, **kwargs)

    def count_nonzero(self, *args, **kwargs):
        return np.count_nonzero(self.values, *args, **kwargs)

    @property
    def shape(self):
        return self.values.shape

    @property
    def __array_interface__(self):
        return self.values.__array_interface__

    def __array__(self):
        return self.values

    def __array_ufunc__(self, *args, **kwargs):
        if args[1] != '__call__':
            # This is a generic catch-all for ufuncs that are not being applied with '__call__' (e.g., operations returning a scalar like 'np.sum()' use reduce instead)
            args = [(x if x is not self else self.values) for x in args]
            kwargs = {k: v if v is not self else self.values for k, v in kwargs.items()}
            return self.values.__array_ufunc__(*args, **kwargs)

        args = [(x if x is not self else self.values) for x in args]
        if 'out' in kwargs and kwargs['out'][0] is self:
            # In-place operations like += use this branch
            kwargs['out'] = self.values
            args[0](*args[2:], **kwargs)
            return self
        else:
            out = args[0](*args[2:], **kwargs)
            if isinstance(out, FusedArray):
                # For some operations (e.g., those involving two FusedArrays) the result of the ufunc will already be a FusedArray
                # In particular, operating on two states will result in a FusedArray where the references to the original People uid_map and uids
                # are still intact. In such cases, we can return the resulting FusedArray directly
                return out
            else:
                # Otherwise, if the result of the ufunc is an array (e.g., because one of the arguments was an array) then
                # we need to wrap it up in a new FusedArray. With '__call__' the dimensions should hopefully be the same and we should
                # be able to reuse the UID arrays directly
                return FusedArray(values=out, uid=self.uid, uid_map=self._uid_map)

    def __array_wrap__(self, out_arr, context=None):
        # This allows numpy operations addition etc. to return instances of FusedArray
        if out_arr.ndim == 0:
            return out_arr.item()
        return FusedArray(values=out_arr, uid=self.uid, uid_map=self._uid_map) # Hardcoding class means State can inherit from FusedArray but return FusedArray base instances


class DynamicView(NDArrayOperatorsMixin):
    def __init__(self, dtype, fill_value=None):
        """
        Args:
            name: name of the result as used in the model
            dtype: datatype
            fill_value: default value for this state upon model initialization. If not provided, it will use the default value for the dtype
            shape: If not none, set to match a string in `pars` containing the dimensionality
            label: text used to construct labels for the result for displaying on plots and other outputs
        """
        self.fill_value = fill_value if fill_value is not None else dtype()
        self.n = 0  # Number of agents currently in use
        self._data = np.empty(0, dtype=dtype)  # The underlying memory array (length at least equal to n)
        self._view = None  # The view corresponding to what is actually accessible (length equal to n)
        self._map_arrays()
        return

    @property
    def _s(self):
        # Return the size of the underlying array (maximum number of agents that can be stored without reallocation)
        return len(self._data)

    @property
    def dtype(self):
        # The specified dtype and the underlying array dtype can be different. For instance, the user might pass in
        # DynamicView(dtype=int) but the underlying array's dtype will be np.dtype('int32'). This distinction is important
        # because the numpy dtype has attributes like 'kind' that the input dtype may not have. We need the DynamicView's
        # dtype to match that of the underlying array so that it can be more seamlessly exchanged with direct numpy arrays
        # Therefore, we retain the original dtype in DynamicView._dtype() and use
        return self._data.dtype

    def __len__(self):
        # Return the number of active elements
        return self.n

    def __repr__(self):
        # Print out the numpy view directly
        return self._view.__repr__()

    def grow(self, n):
        # If the total number of agents exceeds the array size, extend the underlying arrays
        if self.n + n > self._s:
            n_new = max(n, int(self._s / 2))  # Minimum 50% growth
            self._data = np.concatenate([self._data, np.full(n_new, dtype=self.dtype, fill_value=self.fill_value)], axis=0)
        self.n += n  # Increase the count of the number of agents by `n` (the requested number of new agents)
        self._map_arrays()

    def _trim(self, inds):
        # Keep only specified indices
        # Note that these are indices, not UIDs!
        n = len(inds)
        self._data[:n] = self._data[inds]
        self._data[n:self.n] = self.fill_value
        self.n = n
        self._map_arrays()

    def _map_arrays(self):
        """
        Set main simulation attributes to be views of the underlying data

        This method should be called whenever the number of agents required changes
        (regardless of whether or not the underlying arrays have been resized)
        """
        self._view = self._data[:self.n]

    def __getitem__(self, key):
        return self._view.__getitem__(key)

    def __setitem__(self, key, value):
        self._view.__setitem__(key, value)

    @property
    def __array_interface__(self):
        return self._view.__array_interface__

    def __array__(self):
        return self._view

    def __array_ufunc__(self, *args, **kwargs):
        args = [(x if x is not self else self._view) for x in args]
        kwargs = {k: v if v is not self else self._view for k, v in kwargs.items()}
        return self._view.__array_ufunc__(*args, **kwargs)


class State(FusedArray):

    def __init__(self, name, dtype, fill_value=None, label=None):
        """

        :param name: A string name for the state
        :param dtype: The dtype to use for this instance
        :param fill_value: Specify default value for new agents. This can be
            - A scalar with the same dtype (or castable to the same dtype) as the State
            - A callable, with a single argument for the number of values to produce
            - An ss.ScipyDistribution instance
        :param label:
        """

        super().__init__(values=None, uid=None, uid_map=None)  # Call the FusedArray constructor

        self.fill_value = fill_value

        self._data = DynamicView(dtype=dtype)
        self.name = name
        self.label = label or name
        self.values = self._data._view
        self._initialized = False

    def __repr__(self):
        if not self._initialized:
            return f'<State {self.name} (uninitialized)>'
        else:
            return FusedArray.__repr__(self)

    def _new_vals(self, uids):
        if isinstance(self.fill_value, ScipyDistribution):
            new_vals = self.fill_value.rvs(uids)
        elif callable(self.fill_value):
            new_vals = self.fill_value(len(uids))
        else:
            new_vals = self.fill_value
        return new_vals

    def initialize(self, sim=None, people=None):
        if self._initialized:
            return

        if sim is not None and people is None:
            people = sim.people

        sim_still_needed = False
        if isinstance(self.fill_value, rv_frozen):
            if sim is not None:
                self.fill_value = ScipyDistribution(self.fill_value, f'{self.__class__.__name__}_{self.label}')
                self.fill_value.initialize(sim, self)
            else:
                sim_still_needed = True

        people.add_state(self, die=False) # CK: should not be needed
        if not sim_still_needed:
            self._uid_map = people._uid_map
            self.uid = people.uid
            self._data.grow(len(self.uid))
            self._data[:len(self.uid)] = self._new_vals(self.uid)
            self.values = self._data._view
            self._initialized = True
        return

    def grow(self, uids):
        """
        Add state for new agents

        This method is normally only called via `People.grow()`.

        :param uids: Numpy array of UIDs for the new agents being added This array should have length n
        """

        n = len(uids)
        self._data.grow(n)
        self.values = self._data._view
        self._data[-n:] = self._new_vals(uids)
        return

    def _trim(self, inds):
        # Trim arrays to remove agents - should only be called via `People.remove()`
        self._data._trim(inds)
        self.values = self._data._view
        return



"""
Test the RNGContainer object from random.py
"""

# %% Imports and settings
import numpy as np
import sciris as sc
import starsim as ss
from starsim.random import NotInitializedException, SeedRepeatException, RepeatNameException
from starsim.random import SingleRNG, MultiRNG
import scipy.stats as sps
import pytest


# %% Define the tests

@pytest.fixture
def rng_container():
    rng_container = ss.RNGContainer()
    rng_container.initialize(base_seed=10)
    return rng_container

@pytest.fixture
def rngs():
    return [ss.RNG('rng0'), ss.RNG('rng1')]

@pytest.fixture
def dists():
    names = ['rng1', 'rng2']
    dists = []
    for rng_name in names:
        d = ss.ScipyDistribution(sps.uniform(), rng=rng_name)
        dists.append(d)
    return dists


def test_rng(rng_container, rngs, n=5):
    """ Simple sample from rng """
    sc.heading('test_container: Testing RNGContainer object')

    rng0 = rngs[0]
    rng0.initialize(rng_container, slots=n)

    draws = rng0.random(n)
    print(f'Created seed and sampled: {draws}')

    assert len(draws) == n
    return draws


def test_dists(rng_container, dists, n=5):
    """ Simple sample from distribution """
    sc.heading('test_container: Testing RNGContainer object')

    d0 = dists[0]
    if isinstance(d0.rng, (SingleRNG, MultiRNG)):
        d0.rng.initialize(rng_container, slots=n)

    uids = np.arange(0,n,2) # every other to make it interesting
    draws = d0.rvs(uids)
    print(f'Created seed and sampled: {draws}')

    assert len(draws) == len(uids)
    return draws


def test_seed(rng_container, rngs, n=5):
    """ Test assignment of seeds """
    if not ss.options.multirng:
        pytest.skip('Seed assignment is only for multirng.')

    sc.heading('test_seed: Testing assignment of seeds')

    rng0, rng1 = rngs
    rng0.initialize(rng_container, slots=n)
    rng1.initialize(rng_container, slots=n)

    print(f'Random generators rng0 and rng1 were assigned seeds {rng0.seed} and {rng1.seed}, respectively')

    assert rng1.seed != rng0.seed
    return rng0, rng1


def test_reset(rng_container, dists, n=5):
    """ Sample, reset, sample """
    sc.heading('test_reset: Testing sample, reset, sample')

    d0 = dists[0]
    if ss.options.multirng:
        d0.rng.initialize(rng_container, slots=n)

    uids = np.arange(0,n,2) # every other to make it interesting
    s_before = d0.rvs(uids)
    rng_container.reset() # Return to step 0
    s_after = d0.rvs(uids)

    print(f'Initial sample', s_before)
    print('Reset')
    print(f'After reset sample', s_after)

    if ss.options.multirng:
        assert np.array_equal(s_after, s_before)
    else:
        assert not np.array_equal(s_after, s_before)
    return s_before, s_after


def test_step(rng_container, dists, n=5):
    """ Sample, step, sample """
    sc.heading('test_step: Testing sample, step, sample')

    d0 = dists[0]
    if ss.options.multirng:
        d0.rng.initialize(rng_container, slots=n)

    uids = np.arange(0,n,2) # every other to make it interesting
    s_before = d0.rvs(uids)
    rng_container.step(10) # 10 steps
    s_after = d0.rvs(uids)

    print(f'Initial sample', s_before)
    print('Step')
    print(f'Subsequent sample', s_after)

    assert not np.array_equal(s_after, s_before)
    return s_before, s_after


def test_initialize(rngs, n=5):
    """ Sample without initializing, should raise exception """
    sc.heading('test_initialize: Testing without initializing, should raise exception.')
    rng_container = ss.RNGContainer()
    #rng_container.initialize(base_seed=3) # Do not initialize

    with pytest.raises(NotInitializedException):
        rngs[0].initialize(rng_container, slots=n)
    return rngs[0]


def test_seedrepeat(rng_container, n=5):
    """ Two random number generators with the same seed, should raise exception """
    sc.heading('test_seedrepeat: Testing two random number generators with the same seed, should raise exception.')

    rng0 = ss.MultiRNG('rng0', seed_offset=0)
    rng0.initialize(rng_container, slots=n)

    with pytest.raises(SeedRepeatException):
        rng1 = ss.MultiRNG('rng1', seed_offset=0)
        rng1.initialize(rng_container, slots=n)
    return rng0, rng1


def test_samplingorder(rng_container, dists, n=5):
    """ Ensure sampling from one RNG doesn't affect another """
    sc.heading('test_samplingorder: Testing from multiple random number generators to test if sampling order matters')

    uids = np.arange(0,n,2) # every other to make it interesting

    d0, d1 = dists
    if ss.options.multirng:
        d0.rng.initialize(rng_container, slots=n)
        d1.rng.initialize(rng_container, slots=n)

    s_before = d0.rvs(uids)
    _ = d1.rvs(uids)

    rng_container.reset()

    _ = d1.rvs(uids)
    s_after = d0.rvs(uids)

    print(f'When sampling rng0 before rng1:', s_before)
    print('Reset')
    print(f'When sampling rng0 after rng1:', s_after)

    if ss.options.multirng:
        assert np.array_equal(s_before, s_after)
    else:
        assert not np.array_equal(s_before, s_after)

    return s_before, s_after


def test_repeatname(rng_container, n=5):
    """ Test two random number generators with the same name """
    sc.heading('test_repeatname: Testing if two random number generators with the same name are allowed')

    rng0 = ss.RNG('test')
    rng0.initialize(rng_container, slots=n)

    rng1 = ss.RNG('test')
    with pytest.raises(RepeatNameException):
        rng1.initialize(rng_container, slots=n)
    return rng0, rng1



'''
Version and license information.
'''

__all__ = ['__version__', '__versiondate__', '__license__']

__version__ = '0.1.5'
__versiondate__ = '2024-01-23'
__license__ = f'STIsim {__version__} ({__versiondate__}) — © 2024 by IDM'



"""
Define example disease modules
"""

import numpy as np
import starsim as ss
import sciris as sc
from .disease import Disease
import scipy.stats as sps

class SIR(Disease):
    """
    Example SIR model

    This class implements a basic SIR model with states for susceptible,
    infected/infectious, and recovered. It also includes deaths, and basic
    results.

    Note that this class is not fully compatible with common random numbers.
    """

    def __init__(self, pars=None, *args, **kwargs):
        default_pars = {
            #'dur_inf': sps.weibull_min(c=lambda self, sim, uids: sim.people.age[uids], scale=10),#, seed='Duration of SIR Infection'),
            #'dur_inf': sps.norm(loc=lambda self, sim, uids: sim.people.age[uids], scale=2),
            'dur_inf': sps.lognorm(s=1, loc=10),
            'seed_infections': sps.bernoulli(p=0.1),
            'death_given_infection': sps.bernoulli(p=0.2),
            'beta': None,
        }

        super().__init__(pars=ss.omerge(default_pars, pars), *args, **kwargs)

        self.susceptible = ss.State('susceptible', bool, True)
        self.infected = ss.State('infected', bool, False)
        self.recovered = ss.State('recovered', bool, False)
        self.t_infected = ss.State('t_infected', float, np.nan)
        self.t_recovered = ss.State('t_recovered', float, np.nan)
        self.t_dead = ss.State('t_dead', float, np.nan)

        return

    def init_results(self, sim):
        """
        Initialize results
        """
        super().init_results(sim)
        self.results += ss.Result(self.name, 'prevalence', sim.npts, dtype=float)
        self.results += ss.Result(self.name, 'new_infections', sim.npts, dtype=int)
        return

    def update_pre(self, sim):
        # Progress infectious -> recovered
        recovered = ss.true(self.infected & (self.t_recovered <= sim.year))
        self.infected[recovered] = False
        self.recovered[recovered] = True

        # Trigger deaths
        deaths = ss.true(self.t_dead <= sim.year)
        if len(deaths):
            sim.people.request_death(deaths)
        return len(deaths)

    def update_death(self, sim, uids):
        # Reset infected/recovered flags for dead agents
        # This is an optional step. Implementing this function means that in `SIR.update_results()` the prevalence
        # calculation does not need to filter the infected agents by the alive agents. An alternative would be
        # to omit implementing this function, and instead filter by the alive agents when calculating prevalence
        super().update_death(sim, uids)
        self.infected[uids] = False
        self.recovered[uids] = False
        return

    def validate_pars(self, sim):
        if self.pars.beta is None:
            self.pars.beta = sc.objdict({k: 1 for k in sim.people.networks})
        return

    def set_initial_states(self, sim):
        """
        Set initial values for states. This could involve passing in a full set of initial conditions,
        or using init_prev, or other. Note that this is different to initialization of the State objects
        i.e., creating their dynamic array, linking them to a People instance. That should have already
        taken place by the time this method is called.
        """
        alive_uids = ss.true(sim.people.alive)
        initial_cases = self.pars['seed_infections'].filter(alive_uids)
        self.infect(sim, initial_cases, None)
        return

    def infect(self, sim, uids, from_uids):
        super().set_prognoses(sim, uids, from_uids)

        # Carry out state changes associated with infection
        self.susceptible[uids] = False
        self.infected[uids] = True
        self.t_infected[uids] = sim.year

        # Calculate and schedule future outcomes for recovery/death
        dur_inf = self.pars['dur_inf'].rvs(uids)
        will_die = self.pars['death_given_infection'].rvs(uids)
        self.t_recovered[uids[~will_die]] = sim.year + dur_inf[~will_die]
        self.t_dead[uids[will_die]] = sim.year + dur_inf[will_die]

        # Update result count of new infections - important to use += because
        # infect() may be called multiple times per timestep
        self.results['new_infections'][sim.ti] += len(uids)
        return

    def make_new_cases(self, sim): # TODO: Use function from STI
        for k, layer in sim.people.networks.items():
            if k in self.pars['beta']:
                rel_trans = (self.infected & sim.people.alive).astype(float)
                rel_sus = (self.susceptible & sim.people.alive).astype(float)
                for a, b, beta in [[layer.contacts['p1'], layer.contacts['p2'], self.pars['beta'][k]],
                                   [layer.contacts['p2'], layer.contacts['p1'], self.pars['beta'][k]]]:
                    # probability of a->b transmission
                    p_transmit = rel_trans[a] * rel_sus[b] * layer.contacts['beta'] * beta * sim.dt
                    new_cases = np.random.random(len(a)) < p_transmit # As this class is not common-random-number safe anyway, calling np.random is perfectly fine!
                    if new_cases.any():
                        self.infect(sim, b[new_cases], a[new_cases])
        return

    def update_results(self, sim):
        super().update_results(sim)
        self.results['prevalence'][sim.ti] = self.results.n_infected[sim.ti] / np.count_nonzero(sim.people.alive)
        return


class NCD(Disease):
    """
    Example non-communicable disease

    This class implements a basic NCD model with risk of developing a condition
    (e.g., hypertension, diabetes), a state for having the condition, and associated
    mortality.
    """
    def __init__(self, pars=None):
        default_pars = {
            'initial_risk': sps.bernoulli(p=0.3), # Initial prevalence of risk factors
            #'affection_rate': ss.rate(p=0.1), # Instantaneous rate of acquisition applied to those at risk (units are acquisitions / year)
            'dur_risk': sps.expon(scale=10),
            'prognosis': sps.weibull_min(c=2, scale=5), # Time in years between first becoming affected and death
        }

        super().__init__(ss.omerge(default_pars, pars))
        self.at_risk      = ss.State('at_risk', bool, False)
        self.affected     = ss.State('affected', bool, False)
        self.ti_affected  = ss.State('ti_affected', int, ss.INT_NAN)
        self.ti_dead      = ss.State('ti_dead', int, ss.INT_NAN)
        return

    @property
    def not_at_risk(self):
        return ~self.at_risk

    def set_initial_states(self, sim):
        """
        Set initial values for states. This could involve passing in a full set of initial conditions,
        or using init_prev, or other. Note that this is different to initialization of the State objects
        i.e., creating their dynamic array, linking them to a People instance. That should have already
        taken place by the time this method is called.
        """
        alive_uids = ss.true(sim.people.alive)
        initial_risk = self.pars['initial_risk'].filter(alive_uids)
        self.at_risk[initial_risk] = True
        self.ti_affected[initial_risk] = sim.ti + sc.randround(self.pars['dur_risk'].rvs(initial_risk) / sim.dt)

        return initial_risk

    def update_pre(self, sim):
        deaths = ss.true(self.ti_dead == sim.ti)
        sim.people.request_death(deaths)
        self.log.add_data(deaths, died=True)
        self.results.new_deaths[sim.ti] = len(deaths) # Log deaths attributable to this module
        return

    def make_new_cases(self, sim):
        #atrisk_uids = ss.true(self.at_risk)
        #new_cases = self.pars['affection_rate'].filter(atrisk_uids)
        new_cases = ss.true(self.ti_affected == sim.ti)
        self.affected[new_cases] = True
        prog_years = self.pars['prognosis'].rvs(new_cases)
        self.ti_dead[new_cases] = sim.ti + sc.randround(prog_years / sim.dt)
        super().set_prognoses(sim, new_cases)
        return new_cases

    def init_results(self, sim):
        """
        Initialize results
        """
        super().init_results(sim)
        self.results += ss.Result(self.name, 'n_not_at_risk', sim.npts, dtype=int)
        self.results += ss.Result(self.name, 'prevalence', sim.npts, dtype=float)
        self.results += ss.Result(self.name, 'new_deaths', sim.npts, dtype=int)
        return

    def update_results(self, sim):
        super().update_results(sim)
        self.results['n_not_at_risk'][sim.ti] = np.count_nonzero(self.not_at_risk & sim.people.alive)
        self.results['prevalence'][sim.ti] = np.count_nonzero(self.affected & sim.people.alive)/np.count_nonzero(sim.people.alive)
        self.results['new_deaths'][sim.ti] = np.count_nonzero(self.ti_dead == sim.ti)
        return



import hashlib
import numpy as np
import starsim as ss

__all__ = ['RNGContainer', 'MultiRNG', 'SingleRNG', 'RNG']


class RNGContainer:
    """
    Class for managing a collection MultiRNG random number generators
    """

    def __init__(self):
        self._rngs = ss.ndict()
        self.used_seed_offsets = set()
        self.initialized = False
        return

    def initialize(self, base_seed):
        self.base_seed = base_seed
        self.initialized = True
        return
    
    def add(self, rng, check_repeats=True):
        """
        Add a random number generator
        
        Can request an offset, will check for overlap
        Otherwise, return value will be used as the seed offset for this rng
        """
        if not self.initialized:
            raise NotInitializedException()

        if rng.name in self._rngs:
            raise RepeatNameException(rng.name)

        if check_repeats:
            if rng.seed_offset in self.used_seed_offsets:
                raise SeedRepeatException(rng.name, rng.seed_offset)
            self.used_seed_offsets.add(rng.seed_offset)

        self._rngs.append(rng)

        return self.base_seed + rng.seed_offset # Add in the base seed

    def step(self, ti):
        for rng in self._rngs.dict_values():
            rng.step(ti)
        return

    def reset(self):
        for rng in self._rngs.dict_values():
            rng.reset()
        return


class NotInitializedException(Exception):
    "Raised when a random number generator or a RNGContainer object is called when not initialized."
    def __init__(self, obj_name=None):
        if obj_name is None: 
            msg = f'An object is being used without proper initialization.'
        else:
            msg = f'The object named {obj_name} is being used prior to initialization.'
        super().__init__(msg)
        return


class SeedRepeatException(Exception):
    "Raised when two random number generators have the same seed."
    def __init__(self, rng_name, seed_offset):
        msg = f'Requested seed offset {seed_offset} for the random number generator named {rng_name} has already been used.'
        super().__init__(msg)
        return


class RepeatNameException(Exception):
    "Raised when adding a random number generator to a RNGContainer when the rng name has already been used."
    def __init__(self, rng_name):
        msg = f'A random number generator with name {rng_name} has already been added.'
        super().__init__(msg)
        return


def RNG(*args, **kwargs):
    """
    Class to choose a random number generator class.
    
    Returns:
    Numpy RandomState singleton or a MultiRNG: Instance of a random number generator
    """
    if ss.options.multirng:
        rng = MultiRNG(*args, **kwargs)
    else:
        rng = SingleRNG(*args, **kwargs)
    return rng


class MultiRNG(np.random.Generator):
    """
    Class for tracking one random number generators associated with one decision per timestep.

    The main use case is to sample random numbers from various distributions
    that are specific to each agent (per decision and timestep) so as to enable
    variance reduction between simulations through the use of common random
    numbers. For example, the user might create a random number generator called
    rng and ultimately ask for randomly distributed random numbers for agents
    with UIDs 1 and 4:

    >>> import starsim as ss
    >>> import numpy as np
    >>> rng = ss.MultiRNG('Test') # The hashed name determines the seed offset.
    >>> rng.initialize(container=None, slots=5) # In practice, slots will be sim.people.slots. When scalar (for testing), an np.arange will be used.
    >>> uids = np.array([1,4])
    >>> rng.random(uids)
    array([0.88110549, 0.86915719])

    In theory, what this is doing is drawing 5 random numbers and returning the
    draws at positions 1 and 4.

    In practice, using UIDs as "slots" (the indices into the larger draw) falls
    apart when new agents are born.  The issue is that one simulation might have
    more births than another, so an agent born in one simulation may not
    get the same UID as that same agent in a comparison simulation.
    
    The solution applied here is for each agent to have a property called "slot"
    that is precisely the index used when selecting from an array of random
    numbers.  When new agents are born, the mother uses her UID to sample a
    random integer for the newborn that is used as the "slot".  With this
    approach, newborns will be identical between two different simulations,
    unless an intervention mechanistically drove a change.

    The slot-based approach is not without challenges.
    * Two newborn agents may received the same "slot," and thus will receive the
      same random draws.
    * The chance of overlapping slots can be reduced by
      allowing mothers to choose from a larger number of possible slots (say up
      to one-million). However, as slots are used as indices, the number of
      random variables drawn for each query must number the maximum slot. So if
      one agent has a slot of 1M, then 1M random numbers will be drawn,
      consuming more time than would be necessary if the maximum slot was
      smaller.
    * The maximum slot is now determined by a new configure parameter named
      "slot_scale". A value of 5 will mean that new agents will be assigned
      slots between 1*N and 5*N, where N is sim.pars['n_agents'].
    """
    
    def __init__(self, name, seed_offset=None, **kwargs):
        """
        Create a random number generator

        seed_offset will be automatically assigned (based on hashing the name) if None
        
        name: a name for this random number generator, like "coin_flip"
        """

        self.name = name
        self.kwargs = kwargs

        if seed_offset is None:
            # Obtain the seed offset by hashing the class name. Don't use python's hash because it is randomized.
            self.seed_offset = int(hashlib.sha256(self.name.encode('utf-8')).hexdigest(), 16) % 10**8
        else:
            # Use user-provided seed_offset (unlikely)
            self.seed_offset = seed_offset

        self.seed = None # Will be determined once added to the RNG Container
        self.initialized = False
        self.ready = True
        return

    def initialize(self, container, slots):
        if self.initialized:
            return

        if container is not None:
            self.seed = container.add(self) # base_seed + seed_offset
        else:
            # Enable use of MultiRNG without a container
            self.seed = self.seed_offset

        if isinstance(slots, int):
            # Handle edge case in which the user wants n sequential slots, as used in testing.
            self.slots = np.arange(slots)
        else:
            self.slots = slots # E.g. sim.people.slots (instead of using uid as the slots directly)

        if 'bit_generator' not in self.kwargs:
            self.kwargs['bit_generator'] = np.random.PCG64DXSM(seed=self.seed)
        super().__init__(**self.kwargs)

        self._init_state = self.bit_generator.state # Store the initial state

        self.initialized = True
        self.ready = True
        return

    def reset(self):
        """ Restore initial state """
        self.bit_generator.state = self._init_state
        self.ready = True
        return

    def step(self, ti):
        """ Advance to time ti step by jumping """
        
        # First reset back to the initial state
        self.reset()

        # Now take ti jumps
        # jumped returns a new bit_generator, use directly instead of setting state?
        self.bit_generator.state = self.bit_generator.jumped(jumps=ti).state
        return


class SingleRNG(np.random.Generator):
    """
    Dummy class that mirrors MultiRNG but provides the numpy singleton rng instead
    """
    
    def __init__(self, name, seed_offset=None, **kwargs):
        """
        Create a random number generator

        seed_offset will be automatically assigned (based on hashing the name) if None
        
        name: a name for this random number generator, like "coin_flip"
        """

        self.name = name
        self.kwargs = kwargs

        if seed_offset is None:
            # Obtain the seed offset by hashing the class name. Don't use python's hash because it is randomized.
            self.seed_offset = int(hashlib.sha256(self.name.encode('utf-8')).hexdigest(), 16) % 10**8
        else:
            # Use user-provided seed_offset (unlikely)
            self.seed_offset = seed_offset

        self.initialized = False
        return

    def initialize(self, container, slots=None):
        if self.initialized:
            return

        if container is not None:
            container.add(self) # base_seed + seed_offset

        if 'bit_generator' not in self.kwargs:
            self.kwargs['bit_generator'] = np.random.mtrand._rand._bit_generator
        super().__init__(**self.kwargs)

        self.initialized = True
        return

    def reset(self):
        pass

    def step(self, ti):
        pass



"""
Test Distributions from distributions.py
"""

# %% Imports and settings
import numpy as np
import sciris as sc
import starsim as ss
from starsim.random import RNG
import scipy.stats as sps
from starsim.distributions import ScipyDistribution
import pytest
import matplotlib.pyplot as plt

@pytest.fixture(params=[5, 50])
def n(request):
    yield request.param


# %% Define the tests
def test_basic():
    dist = sps.norm(loc=1, scale=1) # Make a distribution
    d = ScipyDistribution(dist)

    sample = d.rvs(1)  # Draw a sample
    samples = d.rvs(10) # Draw several samples

    if ss.options.multirng:
        print('Attempting to sample for specific UIDs, but have not provided a MultiRNG so an Exception should be raised.')
        with pytest.raises(Exception):
            d.rvs(np.array([1,3,8])) # Draw samples for specific uids
    else:
        d.rvs(np.array([1,3,8])) # Draw three samples

    mu = 5 # mean (mu) of the underlying normal
    sigma = 1 # stdev (sigma) of the underlying normal
    s = sigma 
    loc = 0
    scale = np.exp(mu)
    fig, ax = plt.subplots(1, 1)
    x = np.linspace(sps.lognorm.ppf(0.01, s=s, loc=loc, scale=scale), sps.lognorm.ppf(0.99, s=s, loc=loc, scale=scale), 100)
    ax.plot(x, sps.lognorm.pdf(x, s=s, loc=loc, scale=scale), 'r-', lw=5, alpha=0.6, label='lognorm pdf')

    mean, var, skew, kurt = sps.lognorm.stats(s, loc=loc, scale=scale, moments='mvsk')
    print('mean', mean, 'var', var, 'skew', skew, 'kurt', kurt)
    print('calc mean', np.exp(mu + sigma**2/2), 'calc var', (np.exp(sigma**2)-1)*np.exp(2*mu+sigma**2)) # Check against math

    temp_poisson_samples = sps.poisson(mu=2).rvs(10)  # Sample from a temporary distribution
    return sample, samples, temp_poisson_samples

def test_uniform_scalar(n):
    """ Create a uniform distribution """
    sc.heading('test_uniform: Testing uniform with scalar parameters')

    rng = ss.MultiRNG('Uniform')
    rng.initialize(container=None, slots=n)
    dist = sps.uniform(loc=1, scale=4)
    dist.random_state = rng
    d = ScipyDistribution(dist)

    uids = np.array([1,3])
    draws = d.rvs(uids)
    print(f'Uniform sample for uids {uids} returned {draws}')

    assert len(draws) == len(uids)
    return draws

def test_uniform_scalar_str(n):
    """ Create a uniform distribution """
    sc.heading('test_uniform: Testing uniform with scalar parameters')

    dist = sps.uniform(loc=1, scale=4)
    d = ScipyDistribution(dist, 'Uniform') # String here!
    if ss.options.multirng:
        d.rng.initialize(container=None, slots=n) # Only really needed for testing as initializing the distribution will do something similar.

    uids = np.array([1,3])
    draws = d.rvs(uids)
    print(f'Uniform sample for uids {uids} returned {draws}')

    assert len(draws) == len(uids)
    return draws


def test_uniform_callable(n):
    """ Create a uniform distribution """
    sc.heading('test_uniform: Testing uniform with callable parameters')

    sim = ss.Sim().initialize()

    loc = lambda self, sim, uids: sim.people.age[uids] # Low
    scale = 1 # Width, could also be a lambda
    dist = sps.uniform(loc=loc, scale=scale)

    d = ScipyDistribution(dist, 'Uniform')
    d.initialize(sim, context=None)

    uids = np.array([1,3])
    draws = d.rvs(uids)
    print(sim.people.age[uids])
    print(f'Uniform sample for uids {uids} returned {draws}')

    assert len(draws) == len(uids)
    return draws


def test_uniform_array(n):
    """ Create a uniform distribution """
    sc.heading('test_uniform: Testing uniform with a array parameters')

    rng = ss.MultiRNG('Uniform')
    rng.initialize(container=None, slots=n)

    uids = np.array([1, 3])
    loc = np.array([1, 100]) # Low
    scale = np.array([2, 25]) # Width

    dist = sps.uniform(loc=loc, scale=scale)
    dist.random_state = rng

    d = ScipyDistribution(dist)
    draws = d.rvs(uids)
    print(f'Uniform sample for uids {uids} returned {draws}')

    assert len(draws) == len(uids)
    return draws


def test_repeat_slot():
    """ Test behavior of repeated slots """
    sc.heading('test_repeat_slot: Test behavior of repeated slots')

    rng = ss.MultiRNG('Uniform')
    slots = np.array([4,2,3,2,2,3])
    n = len(slots)
    rng.initialize(container=None, slots=slots)

    uids = np.arange(n)
    loc = np.arange(n) # Low
    scale = 1 # Width

    dist = sps.uniform(loc=loc, scale=scale)
    dist.random_state = rng

    d = ScipyDistribution(dist)
    draws = d.rvs(uids)
    print(f'Uniform sample for uids {uids} returned {draws}')

    assert len(draws) == len(uids)

    unique_slots = np.unique(slots)
    for s in unique_slots:
        inds = np.where(slots==s)[0]
        frac, integ = np.modf(draws[inds])
        assert np.allclose(integ, loc[inds]) # Integral part should match the loc
        if ss.options.multirng:
            # Same random numbers, so should be same fractional part
            assert np.allclose(frac, frac[0])
    return draws



# %% Run as a script
if __name__ == '__main__':
    # Start timing
    ###T = sc.tic()

    n = 5
    nTrials = 3

    for multirng in [True, False]:
        ss.options(multirng=multirng)
        sc.heading('Testing with multirng set to', multirng)

        times = []
        for trial in range(nTrials):
            T = sc.tic()

            # Run tests - some will only pass if multirng is True
            test_basic()
            test_uniform_scalar(n)
            test_uniform_scalar_str(n)
            test_uniform_callable(n)
            test_uniform_array(n)

            times.append(sc.toc(T, doprint=False, output=True))

    print(times)
    ###sc.toc(T)

    plt.show()
    print('Done.')


"""Random networks"""

import starsim as ss
import numpy as np
import numba as nb
from .networks import Network
from scipy.stats._distn_infrastructure import rv_frozen
from typing import Union

__all__ = ['RandomNetwork']

class RandomNetwork(Network):

    def __init__(self, *, n_contacts : Union[int, rv_frozen], dynamic=True, **kwargs):
        """
        :param n_contacts: A SciPy distribution from which the number of
        contacts are sampled e.g., sps.nbinom(n=5, p=2). Note distribution
        parameter values can be callable for greater functionality.
        :param dynamic: If True, regenerate contacts each timestep
        """
        super().__init__(**kwargs)
        if isinstance(n_contacts, rv_frozen):
            self.n_contacts = ss.ScipyDistribution(n_contacts, f'{self.__class__.__name__}_{self.name}_{self.label}')

        self.dynamic = dynamic

    def initialize(self, sim):
        super().initialize(sim)
        if isinstance(self.n_contacts, ss.ScipyDistribution):
            self.n_contacts.initialize(sim, self)
        self.update(sim.people, force=True)

    @staticmethod
    @nb.njit
    def get_contacts(inds, number_of_contacts):
        """
        Efficiently generate contacts

        Note that because of the shuffling operation, each person is assigned 2N contacts
        (i.e. if a person has 5 contacts, they appear 5 times in the 'source' array and 5
        times in the 'target' array). Therefore, the `number_of_contacts` argument to this
        function should be HALF of the total contacts a person is expected to have, if both
        the source and target array outputs are used (e.g. for social contacts)

        adjusted_number_of_contacts = np.round(number_of_contacts / 2).astype(cvd.default_int)

        Whereas for asymmetric contacts (e.g. staff-public interactions) it might not be necessary

        Args:
            inds: List/array of person indices
            number_of_contacts: List/array the same length as `inds` with the number of unidirectional
            contacts to assign to each person. Therefore, a person will have on average TWICE this number
            of random contacts.

        Returns: Two arrays, for source and target
        """

        total_number_of_half_edges = np.sum(number_of_contacts)
        count = 0
        source = np.zeros((total_number_of_half_edges,), dtype=ss.int_)
        for i, person_id in enumerate(inds):
            n_contacts = number_of_contacts[i]
            source[count : count + n_contacts] = person_id
            count += n_contacts
        target = np.random.permutation(source)
        return source, target

    def update(self, people: ss.People, force: bool = True) -> None:
        """
        Regenerate contacts

        Args:
            force: If True, ignore the `self.dynamic` flag. This is required for initialization.

        """

        if not self.dynamic and not force:
            return

        if isinstance(self.n_contacts, ss.ScipyDistribution):
            number_of_contacts = self.n_contacts.rvs(people.alive) # or people.uid?
        else:
            number_of_contacts = np.full(len(people), self.n_contacts)

        number_of_contacts = np.round(number_of_contacts / 2).astype(ss.int_)  # One-way contacts
        self.contacts.p1, self.contacts.p2 = self.get_contacts(people.uid.__array__(), number_of_contacts)
        self.contacts.beta = np.ones(len(self.contacts.p1), dtype=ss.float_)


"""
Run simplest tests
"""

# %% Imports and settings
import starsim as ss
import matplotlib.pyplot as plt
import scipy.stats as sps
import numpy as np

def test_sir():
    ppl = ss.People(10000)
    ppl.networks = ss.ndict(ss.RandomNetwork(n_contacts=sps.poisson(mu=4)))

    sir_pars = {
        'dur_inf': sps.norm(loc=10), # Override the default distribution
    }
    sir = ss.SIR(sir_pars)

    # You can also change the parameters of the default lognormal distribution directly!
    #sir.pars['dur_inf'].kwds['loc'] = 5 
    
    # Or why not put a lambda here for fun!
    sir.pars['dur_inf'].kwds['loc'] = lambda self, sim, uids: sim.people.age[uids]/10

    sir.pars['beta'] = {'randomnetwork': 0.1}
    sim = ss.Sim(people=ppl, diseases=sir)
    sim.run()

    # CK: parameters changed
    # assert len(sir.log.out_edges(np.nan)) == sir.pars.initial # Log should match initial infections
    df = sir.log.line_list # Check generation of line-list
    # assert df.source.isna().sum() == sir.pars.initial # Check seed infections in line list

    plt.figure()
    plt.stackplot(
        sim.yearvec,
        sir.results.n_susceptible,
        sir.results.n_infected,
        sir.results.n_recovered,
        sim.results.new_deaths.cumsum(),
    )
    plt.legend(['Susceptible', 'Infected', 'Recovered', 'Dead'])
    plt.xlabel('Year')
    plt.title('SIR')
    return

#@pytest.mark.skip(reason="Haven't converted yet")
def test_ncd():
    ppl = ss.People(10000)
    ppl.networks = None
    ncd = ss.NCD()
    sim = ss.Sim(people=ppl, diseases=ncd)
    sim.run()

    assert len(ncd.log.out_edges) == ncd.log.number_of_edges()
    df = ncd.log.line_list # Check generation of line-list
    assert df.source.isna().all()

    plt.figure()
    plt.stackplot(
        sim.yearvec,
        ncd.results.n_not_at_risk,
        ncd.results.n_at_risk-ncd.results.n_affected,
        ncd.results.n_affected,
        sim.results.new_deaths.cumsum(),
    )
    plt.legend(['Not at risk','At risk','Affected', 'Dead'])
    plt.xlabel('Year')
    plt.title('NCD')
    return


if __name__ == '__main__':
    ss.options(multirng=False)
    sim1 = test_sir()
    sim2 = test_ncd()
    plt.show()



"""
Define analyzers
"""

import starsim as ss


__all__ = ['Analyzer']


class Analyzer(ss.Module):

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

    def update_results(self, sim):
        raise NotImplementedError


"""
Define default syphilis disease module
"""

import numpy as np
import sciris as sc
import pandas as pd
from sciris import randround as rr
import scipy.stats as sps

import starsim as ss
from .disease import STI

__all__ = ['Syphilis']


class Syphilis(STI):

    def __init__(self, pars=None):
        super().__init__(pars)

        # Adult syphilis states
        self.adult_states = ['exposed', 'primary', 'secondary', 'latent_temp', 'latent_long', 'tertiary', 'immune']
        self.exposed = ss.State('exposed', bool, False)  # AKA incubating. Free of symptoms, not transmissible
        self.primary = ss.State('primary', bool, False)  # Primary chancres
        self.secondary = ss.State('secondary', bool, False)  # Inclusive of those who may still have primary chancres
        self.latent_temp = ss.State('latent_temp', bool, False)  # Relapses to secondary (~1y)
        self.latent_long = ss.State('latent_long', bool, False)  # Can progress to tertiary or remain here
        self.tertiary = ss.State('tertiary', bool, False)  # Includes complications (cardio/neuro/disfigurement)
        self.immune = ss.State('immune', bool, False)  # After effective treatment people may acquire temp immunity

        # Congenital syphilis states
        self.congenital = ss.State('congenital', bool, False)

        # Duration of stages
        self.dur_exposed = ss.State('dur_exposed', float, np.nan)
        self.dur_primary = ss.State('dur_primary', float, np.nan)
        self.dur_secondary = ss.State('dur_secondary', float, np.nan)
        self.dur_latent_temp = ss.State('dur_latent_temp', float, np.nan)
        self.dur_latent_long = ss.State('dur_latent_long', float, np.nan)
        self.dur_tertiary = ss.State('dur_tertiary', float, np.nan)
        self.dur_infection = ss.State('dur_infection', float, np.nan)  # Sum of all the stages

        # Timestep of state changes
        self.ti_exposed = ss.State('ti_exposed', int, ss.INT_NAN)
        self.ti_primary = ss.State('ti_primary', int, ss.INT_NAN)
        self.ti_secondary = ss.State('ti_secondary', int, ss.INT_NAN)
        self.ti_latent_temp = ss.State('ti_latent_temp', int, ss.INT_NAN)
        self.ti_latent_long = ss.State('ti_latent_long', int, ss.INT_NAN)
        self.ti_tertiary = ss.State('ti_tertiary', int, ss.INT_NAN)
        self.ti_immune = ss.State('ti_immune', int, ss.INT_NAN)
        self.ti_miscarriage = ss.State('ti_miscarriage', int, ss.INT_NAN)
        self.ti_nnd = ss.State('ti_nnd', int, ss.INT_NAN)
        self.ti_stillborn = ss.State('ti_stillborn', int, ss.INT_NAN)
        self.ti_congenital = ss.State('ti_congenital', int, ss.INT_NAN)

        # Parameters
        default_pars = dict(
            # Adult syphilis natural history, all specified in years
            dur_exposed=ss.lognorm(mean=1 / 12, stdev=1 / 36),  # https://pubmed.ncbi.nlm.nih.gov/9101629/
            dur_primary=ss.lognorm(mean=1.5 / 12, stdev=1 / 36),  # https://pubmed.ncbi.nlm.nih.gov/9101629/
            dur_secondary=sps.norm(loc=3.6 / 12, scale=1.5 / 12),  # https://pubmed.ncbi.nlm.nih.gov/9101629/
            dur_latent_temp=ss.lognorm(mean=1, stdev=6 / 12),  # https://pubmed.ncbi.nlm.nih.gov/9101629/
            dur_latent_long=ss.lognorm(mean=20, stdev=8),  # https://pubmed.ncbi.nlm.nih.gov/9101629/
            p_latent_temp=sps.bernoulli(p=0.25),  # https://pubmed.ncbi.nlm.nih.gov/9101629/
            p_tertiary=sps.bernoulli(p=0.35),  # https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4917057/

            # Congenital syphilis outcomes
            # Birth outcomes coded as:
            #   0: Neonatal death
            #   1: Stillborn
            #   2: Congenital syphilis
            #   3: Live birth without syphilis-related complications
            # Source: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5973824/)
            birth_outcomes=sc.objdict(
                active=sps.rv_discrete(values=([0, 1, 2, 3, 4], [0.125, 0.125, 0.20, 0.35, 0.200])),
                latent=sps.rv_discrete(values=([0, 1, 2, 3, 4], [0.050, 0.075, 0.10, 0.05, 0.725])),
            ),
            birth_outcome_keys=['miscarriage', 'nnd', 'stillborn', 'congenital'],

            # Initial conditions
            seed_infections=sps.bernoulli(p=0.03),
        )
        self.pars = ss.omerge(default_pars, self.pars)

        return

    @property
    def active(self):
        """ Active - only active infections can transmit through sexual contact """
        return self.primary | self.secondary

    @property
    def latent(self):
        """ Latent """
        return self.latent_temp | self.latent_long

    @property
    def infectious(self):
        """ Infectious - includes latent infections, which can transmit vertically but not sexually """
        return self.active | self.latent

    def init_results(self, sim):
        """ Initialize results """
        super().init_results(sim)
        self.results += ss.Result(self.name, 'new_nnds', sim.npts, dtype=int, scale=True)
        self.results += ss.Result(self.name, 'new_stillborns', sim.npts, dtype=int, scale=True)
        self.results += ss.Result(self.name, 'new_congenital', sim.npts, dtype=int, scale=True)
        return

    def update_results(self, sim):
        """ Update results """
        super(Syphilis, self).update_results(sim)
        return

    def update_pre(self, sim):
        """ Updates prior to interventions """

        # Primary
        primary = self.exposed & (self.ti_primary <= sim.ti)
        self.primary[primary] = True
        self.exposed[primary] = False

        # Secondary from primary
        secondary_from_primary = self.primary & (self.ti_secondary <= sim.ti)
        self.secondary[secondary_from_primary] = True
        self.primary[secondary_from_primary] = False
        self.set_secondary_prognoses(sim, ss.true(secondary_from_primary))

        # Secondary reactivation from latent
        secondary_from_latent = self.latent_temp & (self.ti_secondary <= sim.ti)
        self.secondary[secondary_from_latent] = True
        self.latent_temp[secondary_from_latent] = False
        self.set_secondary_prognoses(sim, ss.true(secondary_from_latent))

        # Latent
        latent_temp = self.secondary & (self.ti_latent_temp <= sim.ti)
        self.latent_temp[latent_temp] = True
        self.secondary[latent_temp] = False
        self.set_latent_temp_prognoses(sim, ss.true(latent_temp))

        # Latent long
        latent_long = self.secondary & (self.ti_latent_long <= sim.ti)
        self.latent_long[latent_long] = True
        self.secondary[latent_long] = False
        self.set_latent_long_prognoses(sim, ss.true(latent_long))

        # Tertiary
        tertiary = self.latent_long & (self.ti_tertiary <= sim.ti)
        self.tertiary[tertiary] = True
        self.latent_long[tertiary] = False

        # Congenital syphilis deaths
        nnd = self.ti_nnd == sim.ti
        stillborn = self.ti_stillborn == sim.ti
        sim.people.request_death(nnd)
        sim.people.request_death(stillborn)

        # Congenital syphilis transmission outcomes
        congenital = self.ti_congenital == sim.ti
        self.congenital[congenital] = True
        self.susceptible[congenital] = False

        return

    def update_results(self, sim):
        super(Syphilis, self).update_results(sim)
        self.results['new_nnds'][sim.ti] = np.count_nonzero(self.ti_nnd == sim.ti)
        self.results['new_stillborns'][sim.ti] = np.count_nonzero(self.ti_stillborn == sim.ti)
        self.results['new_congenital'][sim.ti] = np.count_nonzero(self.ti_congenital == sim.ti)
        return

    def make_new_cases(self, sim):
        # TODO: for now, still using generic transmission method, but could replace here if needed
        super(Syphilis, self).make_new_cases(sim)
        return

    def set_prognoses(self, sim, target_uids, source_uids=None):
        """
        Set initial prognoses for adults newly infected with syphilis
        """

        # Subset target_uids to only include ones with active infection
        if source_uids is not None:
            active_sources = self.active[source_uids].values.nonzero()[-1]
            uids = target_uids[active_sources]
        else:
            uids = target_uids

        self.susceptible[uids] = False
        self.exposed[uids] = True
        self.infected[uids] = True
        self.ti_exposed[uids] = sim.ti
        self.ti_infected[uids] = sim.ti

        # Set future dates and probabilities
        # Exposed to primary
        dur_exposed = self.pars.dur_exposed.rvs(uids)
        self.dur_exposed[uids] = dur_exposed
        self.ti_primary[uids] = sim.ti + rr(dur_exposed / sim.dt)
        self.dur_infection[uids] = dur_exposed

        # Primary to secondary
        dur_primary = self.pars.dur_primary.rvs(uids)
        self.ti_secondary[uids] = self.ti_primary[uids] + rr(dur_primary / sim.dt)
        self.dur_infection[uids] += dur_primary

        return

    def set_secondary_prognoses(self, sim, uids):
        """ Set prognoses for people who have just progressed to secondary infection """

        # Secondary to latent_temp or latent_long
        latent_temp_uids = self.pars.p_latent_temp.filter(uids)
        latent_long_uids = np.setdiff1d(uids, latent_temp_uids)

        dur_secondary_temp = self.pars.dur_secondary.rvs(latent_temp_uids)
        self.ti_latent_temp[latent_temp_uids] = self.ti_secondary[latent_temp_uids] + rr(dur_secondary_temp / sim.dt)
        self.dur_infection[latent_temp_uids] += dur_secondary_temp

        dur_secondary_long = self.pars.dur_secondary.rvs(latent_long_uids)
        self.ti_latent_long[latent_long_uids] = self.ti_secondary[latent_long_uids] + rr(dur_secondary_long / sim.dt)
        self.dur_infection[latent_long_uids] += dur_secondary_long

        return

    def set_latent_temp_prognoses(self, sim, uids):
        # Primary to secondary
        dur_latent_temp = self.pars.dur_latent_temp.rvs(uids)
        self.ti_secondary[uids] = self.ti_latent_temp[uids] + rr(dur_latent_temp / sim.dt)
        self.dur_infection[uids] += dur_latent_temp
        return

    def set_latent_long_prognoses(self, sim, uids):
        # Primary to secondary
        dur_latent_long = self.pars.dur_latent_long.rvs(uids)
        self.ti_secondary[uids] = self.ti_latent_temp[uids] + rr(dur_latent_long / sim.dt)
        self.dur_infection[uids] += dur_latent_long

        # Latent_long to tertiary
        tertiary_uids = self.pars.p_tertiary.filter(uids)
        dur_latent_long = self.pars.dur_latent_long.rvs(tertiary_uids)
        self.ti_tertiary[tertiary_uids] = self.ti_latent_long[tertiary_uids] + rr(dur_latent_long / sim.dt)
        self.dur_infection[tertiary_uids] += dur_latent_long

        return

    def set_congenital(self, sim, target_uids, source_uids=None):
        """
        Natural history of syphilis for congenital infection
        """

        # Determine outcomes
        for state in ['active', 'latent']:

            source_state_inds = getattr(self, state)[source_uids].values.nonzero()[-1]
            uids = target_uids[source_state_inds]

            if len(uids) > 0:

                # Birth outcomes must be modified to add probability of susceptible birth
                birth_outcomes = self.pars.birth_outcomes[state]
                assigned_outcomes = birth_outcomes.rvs(uids)-uids  # WHY??
                time_to_birth = -sim.people.age

                # Schedule events
                for oi, outcome in enumerate(self.pars.birth_outcome_keys):
                    o_uids = uids[assigned_outcomes == oi]
                    if len(o_uids) > 0:
                        ti_outcome = f'ti_{outcome}'
                        vals = getattr(self, ti_outcome)
                        vals[o_uids] = sim.ti + rr(time_to_birth[o_uids].values / sim.dt)
                        setattr(self, ti_outcome, vals)

        return




"""
Test the RNG object from random.py
"""

# %% Imports and settings
import numpy as np
import sciris as sc
import starsim as ss
from starsim.random import RNG
import pytest

@pytest.fixture
def rng(request, slots=5, base_seed=1, name='Test', **kwargs):
    return make_rng(slots, base_seed, name, **kwargs)

def make_rng(slots=5, base_seed=1, name='Test', **kwargs):
    rng_container = ss.RNGContainer()
    rng_container.initialize(base_seed=base_seed)
    rng = ss.RNG(name, **kwargs)
    rng.initialize(rng_container, slots=slots)
    return rng


# %% Define the tests
def test_random(rng, n=5):
    """ Simple random draw """
    sc.heading('test_random: Testing simple random draw from a RNG object')
    draws = rng.random(n)
    print(f'Sampled {n} random draws', draws)
    assert len(draws) == n
    return draws


def test_SingleRNG_using_np_singleton(rng, n=5):
    """ Make sure the SingleRNG is using the numpy.random singleton """
    if ss.options.multirng:
        pytest.skip('Skipping multirng mode')
    sc.heading('test_singleton_rng: Testing RNG object')

    np.random.seed(0)
    draws_np = np.random.rand(n)
    np.random.seed(0)
    draws_SingleRNG = rng.random(n)
    assert np.array_equal(draws_np, draws_SingleRNG)
    return draws_np, draws_SingleRNG


def test_reset(rng, n=5):
    """ Sample, reset, sample """
    sc.heading('test_reset: Testing sample, reset, sample')

    draws1 = rng.random(n)
    print(f'Random sample of size {n} returned {draws1}')

    print('Reset')
    rng.reset()

    draws2 = rng.random(n)
    print(f'After reset, random sample of size {n} returned {draws2}')

    if isinstance(rng, ss.MultiRNG):
        assert np.array_equal(draws1, draws2)
    else:
        assert not np.array_equal(draws1, draws2)
    return draws1, draws2


def test_step(rng, n=5):
    """ Sample, step, sample """
    sc.heading('test_step: Testing sample, step, sample')

    draws1 = rng.random(n)
    print(f'Random sample of size {n} returned {draws1}')

    print('Reset')
    rng.step(1)

    draws2 = rng.random(n)
    print(f'After reset, random sample of size {n} returned {draws2}')

    assert not np.array_equal(draws1, draws2)
    return draws1, draws2


def test_seed(n=5):
    """ Changing seeds """
    sc.heading('test_seed: Testing sample with seeds 0 and 1')

    # NOTE: SingleRNG is using the numpy random singleton, and thus ignores base_seed here

    rng0 = make_rng(n, base_seed=0)
    draws0 = rng0.uniform(size=n)
    print(f'Random sample of size {n} for rng0 with base_seed 0 returned {draws0}')

    rng1 = make_rng(n, base_seed=1)
    draws1 = rng1.uniform(size=n)
    print(f'Random sample of size {n} for rng1 with base_seed 1 returned {draws1}')

    assert not np.array_equal(draws0, draws1)
    return draws0, draws1


# %% Run as a script
if __name__ == '__main__':
    # Start timing
    T = sc.tic()

    n=5

    for multirng in [True, False]:
        ss.options(multirng=multirng)
        sc.heading('Testing with multirng set to', multirng)

        # Run tests - some will only pass if multirng is True
        test_random(n)
        test_reset(n)
        test_step(n)
        test_seed(n)

    sc.toc(T)
    print('Done.')



'''
Disease modules
'''

import sciris as sc
import starsim as ss
from scipy.stats._distn_infrastructure import rv_frozen

__all__ = ['Module']


class Module(sc.prettyobj):

    def __init__(self, pars=None, name=None, label=None, requires=None, *args, **kwargs):
        self.pars = ss.omerge(pars)
        self.name = name if name else self.__class__.__name__.lower() # Default name is the class name
        self.label = label if label else ''
        self.requires = sc.mergelists(requires)
        self.results = ss.ndict(type=ss.Result)
        self.initialized = False
        self.finalized = False

        return

    def check_requires(self, sim):
        for req in sc.tolist(self.requires):
            if req not in [m.__class__ for m in sim.modules]:
                raise Exception(f'{self.name} (label={self.label}) requires module {req} but the Sim did not contain a module of this type.')
        return

    def initialize(self, sim):
        """
        Perform initialization steps

        This method is called once, as part of initializing a Sim

        :param sim:
        :return:
        """
        self.check_requires(sim)

        # Connect the random number generators to the sim. The RNGs for this module should be initialized
        # first as some of the module's State instances may require them to generate initial values
        for rng in self.rngs:
            if not rng.initialized:
                rng.initialize(sim.rng_container, sim.people.slot)

        # Initialize distributions in pars
        for key, value in self.pars.items():
            if isinstance(value, rv_frozen):
                self.pars[key] = ss.ScipyDistribution(value, f'{self.name}_{self.label}_{key}')
                self.pars[key].initialize(sim, self)
            #elif isinstance(value, ss.rate):
            #    self.pars[key].initialize(sim, f'{self.name}_{self.label}_{key}')

        for key, value in self.__dict__.items():
            if isinstance(value, rv_frozen):
                setattr(self, key, ss.ScipyDistribution(value, f'{self.name}_{self.label}_{key}'))
                getattr(self, key).initialize(sim, self)

        # Connect the states to the sim
        # Will use random numbers, so do after distribution initialization
        for state in self.states:
            state.initialize(sim)

        self.initialized = True
        return

    def finalize(self, sim):
        self.finalize_results(sim)
        self.finalized = True
        return

    def finalize_results(self, sim):
        """
        Finalize results
        """
        # Scale results
        for reskey, res in self.results.items():
            if isinstance(res, ss.Result) and res.scale:
                self.results[reskey] = self.results[reskey]*sim.pars.pop_scale
        return

    @property
    def states(self):
        """
        Return a flat collection of all states

        The base class returns all states that are contained in top-level attributes
        of the Module. If a Module stores states in a non-standard location (e.g.,
        within a list of states, or otherwise in some other nested structure - perhaps
        due to supporting features like multiple genotypes) then the Module should
        overload this attribute to ensure that all states appear in here.

        :return:
        """
        return [x for x in self.__dict__.values() if isinstance(x, ss.State)]

    @property
    def rngs(self):
        """
        Return a flat collection of all random number generators, as with states above

        :return:
        """
        return [x for x in self.__dict__.values() if isinstance(x, (ss.MultiRNG, ss.SingleRNG))]

    @property
    def scipy_dbns(self):
        """
        Return a flat collection of all ScipyDistributions

        :return:
        """
        return [x for x in self.__dict__.values() if isinstance(x, ss.ScipyDistribution)] \
             + [x for x in self.pars.values()     if isinstance(x, ss.ScipyDistribution)]



"""
Run syphilis
"""

# %% Imports and settings
import numpy as np
import starsim as ss
import pandas as pd
import matplotlib.pyplot as plt
import scipy.stats as sps


def make_syph_sim():
    """ Make a sim with syphilis - used by several subsequent tests """
    syph = ss.Syphilis()
    syph.pars['beta'] = {'mf': [0.95, 0.75], 'maternal': [0.99, 0]}
    syph.pars['seed_infections'] = sps.bernoulli(p=0.1)

    # Make demographic modules
    fertility_rates = {'fertility_rate': pd.read_csv(ss.root / 'tests/test_data/nigeria_asfr.csv')}
    pregnancy = ss.Pregnancy(pars=fertility_rates)
    death_rates = {'death_rate': pd.read_csv(ss.root / 'tests/test_data/nigeria_deaths.csv')}
    death = ss.background_deaths(death_rates)

    # Make people and networks
    ss.set_seed(1)
    ppl = ss.People(500, age_data=pd.read_csv(ss.root / 'tests/test_data/nigeria_age.csv')) # CK: temporary small pop size
    mf = ss.mf(
        pars=dict(duration_dist=ss.lognorm(mean=0.1, stdev=0.5))
    )
    maternal = ss.maternal()
    ppl.networks = ss.ndict(mf, maternal)

    sim_kwargs = dict(
        dt=1/2, # CK: temporary long dt
        total_pop=93963392,
        start=1990,
        n_years=40,
        people=ppl,
        diseases=syph,
        demographics=[pregnancy, death],
    )

    return sim_kwargs


def test_syph():

    sim_kwargs = make_syph_sim()

    class check_states(ss.Analyzer):

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.okay = True
            return

        def update_results(self, sim):
            return self.apply(sim)

        def apply(self, sim):
            """
            Checks states that should be mutually exlusive and collectively exhaustive
            """
            sppl = sim.people.syphilis

            # Infection states: people must be exactly one of these
            s1 = (sppl.susceptible | sppl.exposed | sppl.primary | sppl.secondary | sppl.latent_temp | sppl.latent_long | sppl.tertiary | sppl.congenital).all()
            if not s1:
                raise ValueError('States should be collectively exhaustive but are not.')
            s2 = ~(sppl.susceptible & sppl.exposed & sppl.primary & sppl.secondary & sppl.latent_temp & sppl.latent_long & sppl.tertiary & sppl.congenital).any()
            if not s2:
                raise ValueError('States should be mutually exclusive but are not.')

            checkall = np.array([s1, s2])
            if not checkall.all():
                self.okay = False

            return

    sim = ss.Sim(analyzers=[check_states], **sim_kwargs)
    sim.run()

    # Check plots
    burnin = 10
    pi = int(burnin/sim.dt)

    plt.figure()
    plt.stackplot(
        sim.yearvec[pi:],
        sim.results.syphilis.n_susceptible[pi:],
        sim.results.syphilis.n_congenital[pi:],
        sim.results.syphilis.n_exposed[pi:],
        sim.results.syphilis.n_primary[pi:],
        sim.results.syphilis.n_secondary[pi:],
        (sim.results.syphilis.n_latent_temp[pi:]+sim.results.syphilis.n_latent_long[pi:]),
        sim.results.syphilis.n_tertiary[pi:],
    )
    plt.legend(['Susceptible', 'Congenital', 'Exposed', 'Primary', 'Secondary', 'Latent', 'Tertiary'], loc='lower right')
    plt.show()

    return sim


if __name__ == '__main__':

    sim = test_syph()



"""
Define default HIV disease module and related interventions
"""

import numpy as np
import sciris as sc
import starsim as ss
from .disease import STI
import scipy.stats as sps

__all__ = ['HIV', 'ART', 'CD4_analyzer']


class HIV(STI):

    def __init__(self, pars=None):
        super().__init__(pars)

        self.on_art      = ss.State('on_art', bool, False)
        self.ti_art      = ss.State('ti_art', int, ss.INT_NAN)
        self.cd4         = ss.State('cd4', float, 500)
        self.ti_dead     = ss.State('ti_dead', int, ss.INT_NAN) # Time of HIV-cause death

        self.death_prob_per_dt = sps.bernoulli(p=self.death_prob)

        self.pars = ss.omerge({
            'cd4_min': 100,
            'cd4_max': 500,
            'cd4_rate': 5,
            'seed_infections': sps.bernoulli(p=0.05),
            'eff_condoms': 0.7,
            'art_efficacy': 0.96,
        }, self.pars)

        return

    @staticmethod
    def death_prob(self, sim, uids):
        return 0.05 / (self.pars.cd4_min - self.pars.cd4_max)**2 *  (self.cd4[uids] - self.pars.cd4_max)**2

    def update_pre(self, sim):
        """ Update CD4 """
        self.cd4[sim.people.alive & self.infected & self.on_art] += (self.pars.cd4_max - self.cd4[sim.people.alive & self.infected & self.on_art])/self.pars.cd4_rate
        self.cd4[sim.people.alive & self.infected & ~self.on_art] += (self.pars.cd4_min - self.cd4[sim.people.alive & self.infected & ~self.on_art])/self.pars.cd4_rate

        self.rel_trans[sim.people.alive & self.infected & self.on_art] = 1 - self.pars['art_efficacy']

        can_die = ss.true(sim.people.alive & sim.people.hiv.infected)
        hiv_deaths = self.death_prob_per_dt.filter(can_die)
        
        sim.people.request_death(hiv_deaths)
        self.ti_dead[hiv_deaths] = sim.ti
        return

    def init_results(self, sim):
        """
        Initialize results
        """
        super().init_results(sim)
        self.results += ss.Result(self.name, 'new_deaths', sim.npts, dtype=int)
        return

    def update_results(self, sim):
        super(HIV, self).update_results(sim)
        self.results['new_deaths'][sim.ti] = np.count_nonzero(self.ti_dead == sim.ti)
        return 

    def make_new_cases(self, sim):
        # eff_condoms = sim.pars[self.name]['eff_condoms'] # TODO figure out how to add this
        super().make_new_cases(sim)
        return

    def set_prognoses(self, sim, uids, source_uids=None):
        super().set_prognoses(sim, uids, source_uids)

        self.susceptible[uids] = False
        self.infected[uids] = True
        self.ti_infected[uids] = sim.ti
        return

    def set_congenital(self, sim, target_uids, source_uids):
        return self.set_prognoses(sim, target_uids, source_uids)


# %% HIV-related interventions

class ART(ss.Intervention):

    def __init__(self, t: np.array, coverage: np.array, **kwargs):
        self.requires = HIV
        self.t = sc.promotetoarray(t)
        self.coverage = sc.promotetoarray(coverage)

        super().__init__(**kwargs)

        self.prob_art_at_infection = sps.bernoulli(p=lambda self, sim, uids: np.interp(sim.year, self.t, self.coverage))
        return

    def initialize(self, sim):
        super().initialize(sim)
        self.results += ss.Result(self.name, 'n_art', sim.npts, dtype=int)
        self.initialized = True
        return

    def apply(self, sim):
        if sim.ti < self.t[0]:
            return

        ti_delay = 1 # 1 time step delay TODO
        recently_infected = ss.true((sim.people.hiv.ti_infected == sim.ti-ti_delay) & sim.people.alive)

        n_added = 0
        if len(recently_infected) > 0:
            inds = self.prob_art_at_infection.filter(recently_infected)
            sim.people.hiv.on_art[inds] = True
            sim.people.hiv.ti_art[inds] = sim.ti
            n_added = len(inds)

        # Add result
        self.results['n_art'][sim.ti] = np.count_nonzero(sim.people.alive & sim.people.hiv.on_art)

        return n_added


#%% Analyzers

class CD4_analyzer(ss.Analyzer):

    def __init__(self):
        self.requires = HIV
        self.cd4 = None
        return

    def initialize(self, sim):
        super().initialize(sim)
        self.cd4 = np.zeros((sim.npts, sim.people.n), dtype=int)
        return

    def apply(self, sim):
        self.cd4[sim.t] = sim.people.hiv.cd4
        return



"""
Define interventions
"""

import starsim as ss
import sciris as sc
import numpy as np
import pandas as pd

__all__ = ['Intervention']


class Intervention(ss.Module):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        return

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

    def apply(self, sim, *args, **kwargs):
        raise NotImplementedError

    def finalize(self, sim):
        super().finalize(sim)
        self.finalize_results(sim)




"""
Define default gonorrhea disease module and related interventions
"""

import numpy as np
import starsim as ss
from .disease import STI
import scipy.stats as sps


__all__ = ['Gonorrhea']


class Gonorrhea(STI):

    def __init__(self, pars=None):
        super().__init__(pars)

        # States additional to the default disease states (see base class)
        self.symptomatic = ss.State('symptomatic', bool, False)
        self.ti_clearance = ss.State('ti_clearance', int, ss.INT_NAN)
        self.p_symp = ss.State('p_symp', float, 1)

        # Parameters
        self.pars = ss.omerge({
            'dur_inf_in_days': sps.lognorm(s=0.6, scale=10),  # median of 10 days (IQR 7–15 days) https://sti.bmj.com/content/96/8/556
            'p_symp': sps.bernoulli(p=0.5),  # Share of infections that are symptomatic. Placeholder value
            'p_clear': sps.bernoulli(p=0.2),  # Share of infections that spontaneously clear: https://sti.bmj.com/content/96/8/556
            'seed_infections': sps.bernoulli(p=0.1),
        }, self.pars)

        # Additional states dependent on parameter values, e.g. self.p_symp?
        # These might be useful for connectors to target, e.g. if HIV reduces p_clear

        return

    def init_results(self, sim):
        """
        Initialize results
        """
        super().init_results(sim)
        self.results += ss.Result(self.name, 'n_symptomatic', sim.npts, dtype=int)
        self.results += ss.Result(self.name, 'new_clearances', sim.npts, dtype=int)
        return

    def update_results(self, sim):
        super(Gonorrhea, self).update_results(sim)
        self.results['n_symptomatic'][sim.ti] = np.count_nonzero(self.symptomatic)
        self.results['new_clearances'][sim.ti] = np.count_nonzero(self.ti_clearance == sim.ti)
        return

    def update_pre(self, sim):
        # What if something in here should depend on another module?
        # I guess we could just check for it e.g., 'if HIV in sim.modules' or
        # 'if 'hiv' in sim.people' or something
        # Natural clearance
        clearances = self.ti_clearance <= sim.ti
        self.susceptible[clearances] = True
        self.infected[clearances] = False
        self.symptomatic[clearances] = False
        self.ti_clearance[clearances] = sim.ti

        return
    
    def make_new_cases(self, sim):
        super(Gonorrhea, self).make_new_cases(sim)
        return

    def set_prognoses(self, sim, target_uids, source_uids=None):
        """
        Natural history of gonorrhea for adult infection
        """
        super().set_prognoses(sim, target_uids, source_uids)

        # Set infection status
        self.susceptible[target_uids] = False
        self.infected[target_uids] = True
        self.ti_infected[target_uids] = sim.ti

        # Set infection status
        symp_uids = self.pars.p_symp.filter(target_uids)
        self.symptomatic[symp_uids] = True

        # Set natural clearance
        clear_uids = self.pars.p_clear.filter(target_uids)
        dur = sim.ti + self.pars['dur_inf_in_days'].rvs(clear_uids)/365/sim.pars.dt # Convert from days to years and then adjust for dt
        self.ti_clearance[clear_uids] = dur

        return

    def set_congenital(self, sim, target_uids, source_uids=None):
        pass



from .version import __version__, __versiondate__, __license__
from .settings      import *
from .parameters    import *
from .utils         import *
from .distributions import *
from .states        import *
from .random        import *
from .people        import *
from .modules       import *
from .networks      import *
from .results       import *
from .demographics  import *
from .interventions import *
from .analyzers     import *
from .demographics  import *
from .connectors    import *
from .diseases      import *
from .sim           import *

import sciris as sc
root = sc.thispath(__file__).parent

# Import the version and print the license
if options.verbose:
    print(__license__)


""" 
Distribution support extending scipy with two key functionalities:
1. Callable parameters
2. Ability to use MultiRNG for common random number support
"""

import numpy as np
from starsim.utils import INT_NAN
from starsim.random import SingleRNG, MultiRNG
from starsim import options, int_
from copy import deepcopy
from scipy.stats._discrete_distns import bernoulli_gen
from scipy.stats import rv_histogram

__all__ = ['ScipyDistribution', 'ScipyHistogram']


class ScipyDistribution():
    def __init__(self, gen, rng=None):
        self.gen = None
        class starsim_gen(type(gen.dist)):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.sim = None
                return

            def initialize(self, sim, context):
                self.sim = sim
                self.context = context
                return

            def rvs(self, *args, **kwargs):
                """
                Return a specified number of samples from the distribution
                """

                size = kwargs['size']
                slots = None
                repeat_slot_flag = False
                self.repeat_slot_handling = {}
                # Work out how many samples to draw. If sampling by UID, this depends on the slots assigned to agents.
                if np.isscalar(size):
                    if not isinstance(size, (int, np.int64, int_)):
                        raise Exception('Input "size" must be an integer')
                    if size < 0:
                        raise Exception('Input "size" cannot be negative')
                    elif size == 0:
                        return np.array([], dtype=int) # int dtype allows use as index, e.g. when filtering
                    else:
                        n_samples = size
                elif len(size) == 0:
                    return np.array([], dtype=int)
                elif size.dtype == bool:
                    n_samples = len(size) if options.multirng else size.sum()
                elif size.dtype in [int, np.int64, int_]:
                    if not options.multirng:
                        n_samples = len(size)
                    else:
                        v = size.__array__() # TODO - check if this works without calling __array__()?
                        try:
                            slots = self.random_state.slots[v].__array__()
                            max_slot = slots.max()
                        except AttributeError as e:
                            if not isinstance(self.random_state, MultiRNG):
                                raise Exception('With options.multirng and passing agent UIDs to a distribution, the random_state of the distribution must be a MultiRNG.')
                            else:
                                if not self.random_state.initialized:
                                    raise Exception('The MultiRNG instance must be initialized before use.')
                            raise e

                        if max_slot == INT_NAN:
                            raise Exception('Attempted to sample from an INT_NAN slot')
                        n_samples = max_slot + 1
                else:
                    raise Exception("Unrecognized input type")

                # Now handle distribution arguments
                for pname in [p.name for p in self._param_info()]:
                    if pname in kwargs and callable(kwargs[pname]):
                        kwargs[pname] = kwargs[pname](self.context, self.sim, size)

                    # Now do slotting if MultiRNG
                    if options.multirng and (pname in kwargs) and (not np.isscalar(kwargs[pname])) and (len(kwargs[pname]) != n_samples):
                        # Fill in the blank. The number of UIDs provided is
                        # hopefully consistent with the length of pars
                        # provided, but we need to expand out the pars to be
                        # n_samples in length.
                        if len(kwargs[pname]) not in [len(size), sum(size)]: # Could handle uid and bool separately? len(size) for uid and sum(size) for bool
                            raise Exception('When providing an array of parameters, the length of the parameters must match the number of agents for the selected size (uids).')
                        pars_slots = np.full(n_samples, fill_value=1, dtype=kwargs[pname].dtype) # self.fill_value
                        if slots is not None:
                            if len(slots) != len(np.unique(slots)):
                                # Tricky - repeated slots!
                                if not repeat_slot_flag:
                                    repeat_slot_u, repeat_slot_ind, inv, cnt = np.unique(slots, return_index=True, return_inverse=True, return_counts=True)
                                self.repeat_slot_handling[pname] = kwargs[pname].__array__().copy() # Save full pars for handling later
                                pars_slots[repeat_slot_u] = self.repeat_slot_handling[pname][repeat_slot_ind] # Take first instance of each
                                repeat_slot_flag = True
                            else:
                                pars_slots[slots] = kwargs[pname]
                        else:
                            pars_slots[size] = kwargs[pname]
                        kwargs[pname] = pars_slots

                kwargs['size'] = n_samples
                vals = super().rvs(*args, **kwargs)
                if repeat_slot_flag:
                    # Handle repeated slots
                    repeat_slot_vals = np.full(len(slots), np.nan)
                    repeat_slot_vals[repeat_slot_ind] = vals[repeat_slot_u] # Store results
                    todo_inds = np.where(np.isnan(repeat_slot_vals))[0]

                    if options.verbose > 1 and cnt.max() > 2:
                        print(f'MultiRNG slots are repeated up to {cnt.max()} times.')

                    #repeat_degree = repeat_slot_cnt.max()
                    while len(todo_inds):
                        repeat_slot_u, repeat_slot_ind, inv, cnt = np.unique(slots[todo_inds], return_index=True, return_inverse=True, return_counts=True)
                        cur_inds = todo_inds[repeat_slot_ind] # Absolute positions being filled this pass

                        # Reset RNG, note that ti=0 on initialization and ti+1
                        # there after, including ti=0. Assuming that repeat
                        # slots are not encountered during sim initialization.
                        if self.sim is not None:
                            self.random_state.step(self.sim.ti+1) 
                        else:
                            # Likely from a test? Just reset the random state.
                            self.random_state.reset()

                        for pname in [p.name for p in self._param_info()]:
                            if pname in self.repeat_slot_handling:
                                kwargs_pname = self.repeat_slot_handling[pname][cur_inds]
                                pars_slots = np.full(n_samples, fill_value=1, dtype=kwargs_pname.dtype) # self.fill_value
                                pars_slots[repeat_slot_u] = kwargs_pname # Take first instance of each
                                kwargs[pname] = pars_slots

                        vals = super().rvs(*args, **kwargs) # Draw again for slot repeat
                        #assert np.allclose(slots[cur_inds], repeat_slot_u) # TEMP: Check alignment
                        repeat_slot_vals[cur_inds] = vals[repeat_slot_u]
                        todo_inds = np.where(np.isnan(repeat_slot_vals))[0]

                    vals = repeat_slot_vals
                
                if isinstance(self, bernoulli_gen):
                    vals = vals.astype(bool)

                # _select:
                if not options.multirng or repeat_slot_flag:
                    return vals

                if np.isscalar(size):
                    return vals
                elif size.dtype == bool:
                    return vals[size]
                else:
                    return vals[slots] # slots defined above

        self.rng = self.set_rng(rng, gen)
        self.gen = starsim_gen(name=gen.dist.name, seed=self.rng)(**gen.kwds)
        return

    @staticmethod
    def set_rng(rng, gen):
        # Handle random generators
        ret = gen.random_state # Default
        if options.multirng and rng and (gen.random_state == np.random.mtrand._rand):
            # MultiRNG, rng not none, and the current "random_state" is the
            # numpy global singleton... so let's override
            if isinstance(rng, str):
                ret = MultiRNG(rng) # Crate a new generator with the user-provided string
            elif isinstance(rng, np.random.Generator):
                ret = rng
            else:
                raise Exception(f'The rng must be a string or a np.random.Generator instead of {type(rng)}')
        return ret

    def initialize(self, sim, context):
        # Passing sim and context here allow callables to receive "self" and sim pointers
        self.gen.dist.initialize(sim, context)
        if isinstance(self.rng, (SingleRNG, MultiRNG)):
            self.rng.initialize(sim.rng_container, sim.people.slot)
        return

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            setattr(result, k, deepcopy(v, memo))
        return result
    
    def __getattr__(self, attr):
        # Returns wrapped generator.(attr) if not a property
        if attr == '__getstate__':
            # Must be from pickle, return a callable function that returns None
            return lambda: None
        #elif attr == '__deepcopy__':
        #    return self
        elif attr in ['__setstate__', '__await__']:
            # Must be from pickle, async programming, copy
            return None
        try:
            return self.__getattribute__(attr)
        except Exception:
            try:
                return getattr(self.gen, attr) # .dist?
            except Exception as e:
                errormsg = f'"{attr}" is not a member of this class or the underlying scipy stats class'
                raise Exception(errormsg)

    def filter(self, size, **kwargs):
        return size[self.gen.rvs(size, **kwargs)]


class ScipyHistogram(rv_histogram):
    def __init__(self, *args, rng=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.random_state = self.set_rng(rng)
        return

    def initialize(self, sim, context):
        # Note: context not used here, but maintained for consistency with ScipyDistribution
        # Passing sim and context here allow callables to receive "self" and sim pointers
        if isinstance(self.random_state, (SingleRNG, MultiRNG)):
            self.random_state.initialize(sim.rng_container, sim.people.slot)
        return

    def set_rng(self, rng):
        # Handle random generators
        ret = self.random_state # Default
        if options.multirng and rng and (self.random_state == np.random.mtrand._rand):
            # MultiRNG, rng not none, and the current "random_state" is the
            # numpy global singleton... so let's override
            if isinstance(rng, str):
                ret = MultiRNG(rng) # Crate a new generator with the user-provided string
            elif isinstance(rng, np.random.Generator):
                ret = rng
            else:
                raise Exception(f'The rng must be a string or a np.random.Generator instead of {type(rng)}')
        return ret

    def rvs(self, *args, **kwargs):
        """
        Return a specified number of samples from the distribution
        """

        size = kwargs['size']
        slots = None
        # Work out how many samples to draw. If sampling by UID, this depends on the slots assigned to agents.
        if np.isscalar(size):
            if not isinstance(size, (int, np.int64, int_)):
                raise Exception('Input "size" must be an integer')
            if size < 0:
                raise Exception('Input "size" cannot be negative')
            elif size == 0:
                return np.array([], dtype=int) # int dtype allows use as index, e.g. when filtering
            else:
                n_samples = size
        elif len(size) == 0:
            return np.array([], dtype=int)
        elif size.dtype == bool:
            n_samples = len(size) if options.multirng else size.sum()
        elif size.dtype in [int, np.int64, int_]:
            if not options.multirng:
                n_samples = len(size)
            else:
                v = size.__array__() # TODO - check if this works without calling __array__()?
                try:
                    slots = self.random_state.slots[v].__array__()
                    max_slot = slots.max()
                except AttributeError as e:
                    if not isinstance(self.random_state, MultiRNG):
                        raise Exception('With options.multirng and passing agent UIDs to a distribution, the random_state of the distribution must be a MultiRNG.')
                    else:
                        if not self.random_state.initialized:
                            raise Exception('The MultiRNG instance must be initialized before use.')
                    raise e

                if max_slot == INT_NAN:
                    raise Exception('Attempted to sample from an INT_NAN slot')
                n_samples = max_slot + 1
        else:
            raise Exception("Unrecognized input type")

        kwargs['size'] = n_samples
        vals = super().rvs(*args, **kwargs)
        
        # _select:
        if not options.multirng:
            return vals

        if np.isscalar(size):
            return vals
        elif size.dtype == bool:
            return vals[size]
        else:
            return vals[slots] # slots defined above

'''
from scipy.stats import bernoulli
class rate(ScipyDistribution):
    """
    Exponentially distributed, accounts for dt.
    Assumes the rate is constant over each dt interval.
    """
    def __init__(self, p, rng=None):
        dist = bernoulli(p=p)
        super().__init__(dist, rng)
        self.rate = rate
        self.dt = None
        return

    def initialize(self, sim, rng):
        self.dt = sim.dt
        self.rng = self.set_rng(rng, self.gen)
        super().initialize(sim)
        return

    def sample(self, size=None):
        n_samples, pars = super().sample(size, rate=self.rate)
        prob = 1 - np.exp(-pars['rate'] * self.dt)
        vals = self.rng.random(size=n_samples)
        vals = self._select(vals, size)
        return vals < prob
'''
